//
//  SynthObject.m
//
//  Created by 成田 拓郎 on 10/04/30.
//  Copyright 2010 lovelyshark. All rights reserved.
//

#import "SynthObject.h"


@implementation SynthObject

@synthesize inputWaveObject;
@synthesize inputControlObject;
@synthesize controlLevel;
@synthesize outputBuffer;

//加法性ホワイト ガウス ノイズ(AWGN:additive white Gaussian noise)
+ (float) AWGN
{
	float noise = (float)random() + (float)random() + (float)random() + (float)random()
	+ (float)random() + (float)random() + (float)random() + (float)random()
	+ (float)random() + (float)random() + (float)random() + (float)random();
	noise = (noise / INT32_MAX) - 6.0f;
	if (noise < -1.0f)		noise = -1.0f;
	else if (noise > 1.0f)	noise = 1.0f;
	
	return noise;
}

//ダミーバッファ
+ (float*) dummyBuffer0
{
	static float dummy[WAVEOUT_SIZE];

	static BOOL dummy_init = YES;
	if (dummy_init) {
		float *dest = dummy;
		float *dest_end = dummy + WAVEOUT_SIZE;
		while (dest != dest_end) {
			*dest = 0.0f;
			++dest;
		}
		dummy_init = NO;
	}

	return dummy;
}

+ (float*) dummyBuffer1
{
	static float dummy[WAVEOUT_SIZE];
	
	static BOOL dummy_init = YES;
	if (dummy_init) {
		float *dest = dummy;
		float *dest_end = dummy + WAVEOUT_SIZE;
		while (dest != dest_end) {
			*dest = 1.0f;
			++dest;
		}
		dummy_init = NO;
	}
	
	return dummy;
}

//出力サンプルレート(全体で共有される)
+ (float) outputSampleRate:(float)rate
{
	static float sample_rate = SAMPLE_RATE;

	if (rate > 0.0f) {	//パラメータが0.0f以下の場合、値を変更しない
		sample_rate = rate;
	}

	return sample_rate;
}

//演算開始処理
- (void) synthBeing
{
	if (flag == NO)	//(未計算)初期化済み
		return;
	
	//出力バッファの準備
	float *dest = buffer;
	float *src  = buffer + bufferLength;
	while (dest != outputBuffer) {
		*dest = *src;
		++dest;
		++src;
	}

	flag = NO;
	
	//入力オブジェクトの初期化
	[inputWaveObject synthBeing];
	[inputControlObject synthBeing];
	[self synthBeingInput];
}

//入力オブジェクトの初期化
//(入力オブジェクトを追加した場合、派生クラスでオーバーライドして、
// 入力オブジェクトのsynthBeginを呼び出す)
- (void) synthBeingInput
{
}

//このオブジェクトの演算関数
- (void) synthFunc:(int)length
{
}

//演算チェーンの実行
- (void) synth:(int)length
{
	if (flag)	//演算済み
		return;

	//入力オブジェクトの演算
	[inputWaveObject synth:length];
	[inputControlObject synth:length];
	[self synthInput:length];

	//演算
	[self synthFunc:length];
	bufferLength = length;
	flag = YES;
}

//入力オブジェクトの演算
//(入力オブジェクトを追加した場合、派生クラスでオーバーライドして、
// 入力オブジェクトのsynth:を呼び出す)
- (void)synthInput:(int)length
{
}

- (id) init
{
	if (self = [super init])
	{
		flag = YES;

//		inputWaveObject = nil;
//		inputControlObject = nil;
		self.controlLevel    = 0.0f;

		for (int i=0; i<WAVEOUT_SIZE; ++i) {
			buffer[i] = 0.0f;
		}
		outputBuffer = buffer + WAVEOUT_BASE;
		bufferLength = 0;

		[SynthObject AWGN];
		[SynthObject dummyBuffer0];
		[SynthObject dummyBuffer1];
	}

	return self;
}

- (void) dealloc
{
	[inputWaveObject release];
	[inputControlObject release];

	[super dealloc];
}

@end
