//
//  SynthesizerOutput.m
//  AnalogSynthesizer
//
//  Created by 成田 拓郎 on 10/05/28 - 2012/06/12.
//  Copyright 2010-2012 lovelyshark. All rights reserved.
//

#import "SynthesizerOutput.h"


@implementation SynthesizerOutput

@synthesize keyboard;

- (int)vcoWaveType {
	return synthVCO[0].waveType;
}
- (void)setVcoWaveType:(int)wavetype {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCO[i].waveType = wavetype;
}

- (float)vcoPulseWidth {
	return synthVCO[0].pulseWidth;
}
- (void)setVcoPulseWidth:(float)pulsewidth {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCO[i].pulseWidth = pulsewidth;
}

- (float)vcoControlLevel {
	return synthVCO[0].controlLevel;
}
- (void)setVcoControlLevel:(float)controllevel {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCO[i].controlLevel = controllevel;
}

- (float)vcfCutOffFrequency {
	return synthVCF[0].cutOffFrequency;
}
- (void)setVcfCutOffFrequency:(float)frequency {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCF[i].cutOffFrequency = frequency;
}

- (float)vcfResonance {
	return synthVCF[0].resonance;
}
- (void)setVcfResonance:(float)resonance {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCF[i].resonance = resonance;
}

- (float)vcfControlLevel {
	return synthVCF[0].controlLevel;
}
- (void)setVcfControlLevel:(float)controllevel {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCF[i].controlLevel = controllevel;
}

- (float)vcaControlLevel {
	return synthVCA[0].controlLevel;
}
- (void)setVcaControlLevel:(float)controllevel {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthVCA[i].controlLevel = controllevel;
}

- (float)egParamAttack {
	return synthEG[0].paramAttack;
}
- (void)setEgParamAttack:(float)paramattack {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthEG[i].paramAttack = paramattack;
}

- (float)egParamDecay {
	return synthEG[0].paramDecay;
}
- (void)setEgParamDecay:(float)paramdecay {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthEG[i].paramDecay = paramdecay;
}

- (float)egParamSustain {
	return synthEG[0].paramSustain;
}
- (void)setEgParamSustain:(float)paramsustain {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthEG[i].paramSustain = paramsustain;
}

- (float)egParamRelease {
	return synthEG[0].paramRelease;
}
- (void)setEgParamRelease:(float)paramrelease {
	for (int i=0; i<SYNTH_NUM; ++i)
		synthEG[i].paramRelease = paramrelease;
}

- (int)lfoWaveType {
	return synthLFO.waveType;
}
- (void)setLfoWaveType:(int)wavetype {
	synthLFO.waveType = wavetype;
}

- (float)lfoPulseWidth {
	return synthLFO.pulseWidth;
}
- (void)setLfoPulseWidth:(float)pulsewidth {
	synthLFO.pulseWidth = pulsewidth;
}

- (float)lfoFrequency {
	return synthLFO.frequency;
}
- (void)setLfoFrequency:(float)frequency {
	synthLFO.frequency = frequency;
}

- (float)echoDelayTime {
	return synthEcho.delayTime;
}
- (void)setEchoDelayTime:(float)delaytime {
	synthEcho.delayTime = delaytime;
}

- (float)echoDelayLevel {
	return synthEcho.delayLevel;
}
- (void)setEchoDelayLevel:(float)delaylevel {
	synthEcho.delayLevel = delaylevel;
}

//初期化
- (id)init
{
	self = [super init];
	if (self != nil)
	{
		//NSUserDefaultsオブジェクトを用意
		NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
		// 環境設定を読み込み
		int   vco_wave_type   = [defaults integerForKey:@"vco_wave_type"];
		if (vco_wave_type < 0)
			vco_wave_type = 0;
		else if (vco_wave_type > 5)
			vco_wave_type = 5;
		float vco_pulse_width = [defaults floatForKey:@"vco_pulse_width"];
		if (vco_pulse_width < 0.01f)
			vco_pulse_width = 0.01f;
		else if (vco_pulse_width > 0.99f)
			vco_pulse_width = 0.99f;
		float vco_lfo_level   = [defaults floatForKey:@"vco_lfo_level"];
		if (vco_lfo_level < -1.0f)
			vco_lfo_level = -1.0f;
		else if (vco_lfo_level > 1.0f)
			vco_lfo_level = 1.0f;
		float vcf_cutoff_frequency = [defaults floatForKey:@"vcf_cutoff_frequency"];
		if (vcf_cutoff_frequency < 500.0f)
			vcf_cutoff_frequency = 500.0f;
		else if (vcf_cutoff_frequency > 10000.0f)
			vcf_cutoff_frequency = 10000.0f;
		float vcf_resonance   = [defaults floatForKey:@"vcf_resonance"];
		if (vcf_resonance < 0.25f)
			vcf_resonance = 0.25f;
		else if (vcf_resonance > 2.0f)
			vcf_resonance = 2.0f;
		float vcf_lfo_level   = [defaults floatForKey:@"vcf_lfo_level"];
		if (vcf_lfo_level  < -1.0f)
			vcf_lfo_level = -1.0f;
		else if (vcf_lfo_level > 1.0f)
			vcf_lfo_level = 1.0f;

		// シンセサイザ設定
		SynthVCO *lfo = [[SynthVCO alloc] init];
		synthLFO = lfo;
		lfo.waveType = VCO_SIN_WAVE;
		lfo.frequency = 5.0f;

		for (int i=0; i<SYNTH_NUM; ++i) {
			SynthVCO *vco = [[SynthVCO alloc] init];
			synthVCO[i] = vco;
			SynthVCF *vcf = [[SynthVCF alloc] init];
			synthVCF[i] = vcf;
			SynthVCA *vca = [[SynthVCA alloc] init];
			synthVCA[i] = vca;
			SynthEG *eg = [[SynthEG alloc] init];
			synthEG[i] = eg;

			vco.waveType     = vco_wave_type;
			vco.pulseWidth   = vco_pulse_width;
			vco.controlLevel = vco_lfo_level;

			vcf.cutOffFrequency = vcf_cutoff_frequency;
			vcf.resonance       = vcf_resonance;
			vcf.controlLevel    = vcf_lfo_level;

			vco.inputControlObject = lfo;	// VCOにLFOを接続
			vcf.inputWaveObject    = vco;	// VCFにVCOを接続
			vcf.inputControlObject = lfo;	// VCFにLFOを接続
			vca.inputWaveObject    = vcf;	// VCAにVCFを接続
			vca.inputControlObject = lfo;	// VCAにLFOを接続
			vca.inputEGObject      = eg;	// VCAにEGを接続

			noteNumber[i] = UINT32_MAX;
		}

		SynthMix *mix = [[SynthMix alloc] init];
		synthMix = mix;
		mix.inputWaveObject    = synthVCA[0];
		if (2 <= SYNTH_NUM)
			mix.inputControlObject = synthVCA[1];
		if (3 <= SYNTH_NUM)
			mix.inputWaveObject3 = synthVCA[2];
		if (4 <= SYNTH_NUM)
			mix.inputWaveObject4 = synthVCA[3];
		
		SynthEcho *echo = [[SynthEcho alloc] init];
		synthEcho = echo;
		echo.inputWaveObject = mix;
	}
	return self;
}

// ステレオ＆非インタリーブ
- (void)stereoNonInterleave:(float*)pBuffer
					   ptr1:(AudioSampleType *)ptr1
					   ptr2:(AudioSampleType *)ptr2
					 length:(UInt32)length
{
    float* pEnd = pBuffer + length;
    while (pBuffer != pEnd) {
        float sinewave = (*pBuffer) * 0.25f;
        ++pBuffer;
		
        if (sinewave < -1.0)
            sinewave = -1.0;
        else if (sinewave > 1.0)
            sinewave = 1.0;
        sinewave = sinewave * INT16_MAX;
		//      sinewave = sinewave * (1 << kAudioUnitSampleFractionBits);
		
        *ptr1 = sinewave; ++ptr1;
        *ptr2 = sinewave; ++ptr2;
    }
}

// ステレオ＆インタリーブ
- (void)stereoInterleave:(float*)pBuffer
                     ptr:(AudioSampleType *)ptr
                  length:(UInt32)length
{
    float* pEnd = pBuffer + length;
    while (pBuffer != pEnd) {
        float sinewave = (*pBuffer) * 0.25f;
        ++pBuffer;
		
        if (sinewave < -1.0)
            sinewave = -1.0;
        else if (sinewave > 1.0)
            sinewave = 1.0;
        sinewave = sinewave * INT16_MAX;
		//      sinewave = sinewave * (1 << kAudioUnitSampleFractionBits);
		
        *ptr = sinewave; ++ptr;
        *ptr = sinewave; ++ptr;
    }
}

// モノラル
- (void)monaural:(float*)pBuffer
          ioData:(AudioBufferList *)ioData
          length:(UInt32)length
{
    // オーディオ バッファに書き込み(モノラル)
    for (int i=0; i < ioData->mNumberBuffers; ++i)
    {
        AudioSampleType *ptr = ioData->mBuffers[i].mData;
        for (int j=0; j < length; ++j)
        {
            float sinewave = pBuffer[j] * 0.25f;
			
            if (sinewave < -1.0)
                sinewave = -1.0;
            else if (sinewave > 1.0)
                sinewave = 1.0;
            sinewave = sinewave * INT16_MAX;
			//          sinewave = sinewave * (1 << kAudioUnitSampleFractionBits);
			
            UInt32 channels = ioData->mBuffers[i].mNumberChannels;
            for (int k=0; k < channels; ++k)
            {
                ptr[j * channels + k] = sinewave;
            }
        }
    }
}

//サウンド コールバック関数(クラス メンバ)
-(OSStatus)outputCallback:(AudioUnitRenderActionFlags *)ioActionFlags
			  inTimeStamp:(const AudioTimeStamp *)inTimeStamp
			  inBusNumber:(UInt32)inBusNumber
		   inNumberFrames:(UInt32)inNumberFrames
				   ioData:(AudioBufferList *)ioData
{
	OSStatus err = noErr;

	//	NSLog(@"mNumberBuffers = %i mNumberChannels = %i", ioData->mNumberBuffers, ioData->mBuffers[0].mNumberChannels);

	//キー入力
	UInt32 key = self.keyboard;
	UInt32 scale = 60 - 24;

	//キーOFF判定
	for (int i=0; i<SYNTH_NUM; ++i) {
		if (noteNumber[i] >= 128)
			continue;
		BOOL flag = YES;
		if (scale <= noteNumber[i] && noteNumber[i] < (scale + 25))
			flag = (key & (1 << (noteNumber[i] - scale))) == 0;
		if (flag == NO)
			continue;
		//キーOFFする
		int mode = synthEG[i].outputMode;
		if (mode != EG_SPACE_MODE && mode != EG_RELEASE_MODE)
			synthEG[i].outputMode = EG_RELEASE_MODE;
		noteNumber[i] = UINT32_MAX;
	}

	//キーON判定
	for (int i=0; i<25; ++i) {
		if ((key & (1 << i)) == 0)
			continue;
		int note_number = scale + i;
		//すでに発音中かを確認
		BOOL flag = NO;
		for (int i=0; (i<SYNTH_NUM) && (flag == NO); ++i) {
			flag = (noteNumber[i] == note_number);
		}
		if (flag)
			continue;
		//発音に適したユニットを確認
		int unit_num = 0;
		float unit_score = synthEG[0].score;
		for (int i=1; i<SYNTH_NUM; ++i) {
			float score = synthEG[i].score;
			if (score > unit_score) {
				unit_num = i;
				unit_score = score;
			}
		}

		//発音
		synthEG[unit_num].outputMode = EG_ATTACK_MODE;
		synthVCO[unit_num].frequency = [SynthVCO note2Frequency:note_number];
		noteNumber[unit_num] = note_number;
	}
	
	//シンセサイザ処理

	[synthEcho synthBeing];
	[synthEcho synth:inNumberFrames];
	
	float* pBuffer = synthEcho.outputBuffer;
//	NSLog(@"%f", *pBuffer);
	
	// オーディオ バッファに書き込み
    if (ioData->mNumberBuffers == 2 &&
        ioData->mBuffers[0].mNumberChannels == 1 &&
        ioData->mBuffers[1].mNumberChannels == 1) {
        // ステレオ＆非インタリーブ
        [self stereoNonInterleave:pBuffer
                             ptr1:ioData->mBuffers[0].mData
                             ptr2:ioData->mBuffers[1].mData
                           length:inNumberFrames];
    }
    else if (ioData->mNumberBuffers == 1 &&
			 ioData->mBuffers[0].mNumberChannels == 2) {
        // ステレオ＆インタリーブ
        [self stereoInterleave:pBuffer
                           ptr:ioData->mBuffers[0].mData
                        length:inNumberFrames];
    }
    else {
        // モノラル
        [self monaural:pBuffer
                ioData:ioData
                length:inNumberFrames];
    }
	
	return err;
}

//終了処理
-(void)dealloc
{
	for (int i=0; i<SYNTH_NUM; ++i) {
		[synthVCO[i] release];
		[synthVCF[i] release];
		[synthVCA[i] release];
		[synthEG[i] release];
	}
	[synthLFO release];
	[synthMix release];
	[synthEcho release];
	[super dealloc];
}

@end
